import { statSync } from 'fs';
import { extname, sep, dirname, resolve } from 'path';
import acorn from 'acorn';
import { walk } from 'estree-walker';
import MagicString from 'magic-string';
import { addExtension, createFilter } from 'rollup-pluginutils';

var firstpass = /\b(?:require|module|exports)\b/;
function commonjs() {
	var options = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

	var filter = createFilter(options.include, options.exclude);

	return {
		resolveId: function resolveId(importee, importer) {
			if (importee[0] !== '.') return; // not our problem

			var withExtension = addExtension(importee);
			var resolved = resolve(dirname(importer), withExtension);

			// look for `foo.js`...
			try {
				statSync(resolved);
				return resolved;
			} catch (err) {}

			// ...then `foo/index.js`
			if (importee !== withExtension) {
				try {
					resolved = resolved.replace(/\.js$/, sep + 'index.js');

					statSync(resolved);
					return resolved;
				} catch (err) {}
			}
		},

		transform: function transform(code, id) {
			if (!filter(id)) return null;
			if (extname(id) !== '.js') return null;
			if (!firstpass.test(code)) return null;

			var ast = undefined;

			try {
				ast = acorn.parse(code, {
					ecmaVersion: 6,
					sourceType: 'module'
				});
			} catch (err) {
				err.message += ' in ' + id;
				throw err;
			}

			var magicString = new MagicString(code);

			var required = {};
			var uid = 0;
			var hasCommonJsExports = false;

			// TODO handle shadowed `require` calls
			var depth = 0;

			walk(ast, {
				enter: function enter(node, parent) {
					if (options.sourceMap) {
						magicString.addSourcemapLocation(node.start);
						magicString.addSourcemapLocation(node.end);
					}

					if (/Function/.test(node.type)) {
						depth += 1;
						return;
					}

					if (/^(Import|Export)/.test(node.type)) {
						return;
					}

					// TODO more accurate check
					if (node.type === 'Identifier' && node.name === 'exports' || node.name === 'module') {
						hasCommonJsExports = true;
						return;
					}

					if (node.type !== 'CallExpression') return;
					if (node.callee.name !== 'require') return;
					if (node.arguments.length !== 1 || node.arguments[0].type !== 'Literal') return; // TODO handle these weird cases?

					var source = node.arguments[0].value;

					var existing = required[source];
					var name = undefined;

					if (!existing) {
						name = 'require$$' + uid++;
						required[source] = { source: source, name: name };
					} else {
						name = required[source].name;
					}

					magicString.overwrite(node.start, node.end, name);
				},

				leave: function leave(node) {
					if (/Function/.test(node.type)) depth -= 1;
				}
			});

			var sources = Object.keys(required);

			if (!sources.length && !hasCommonJsExports) return null;

			var importBlock = sources.length ? sources.map(function (source) {
				return 'import ' + required[source].name + ' from \'' + source + '\';';
			}).join('\n') : '';

			var intro = 'var exports = {}, module = { \'exports\': exports };\n';
			var outro = '\nexport default module.exports;';

			magicString.trim().prepend(importBlock + intro).append(outro);

			code = magicString.toString();
			var map = options.sourceMap ? magicString.generateMap() : null;

			return { code: code, map: map };
		}
	};
}

export default commonjs;